/* 
 * SCSIping: use SCSI commands to query a SCSI drive, and suggest two 
 *           format.dat entries to use for the drive.
 * 
 *           Works on Sun4c/sun4m only. 
 *
 *   Author: Poul-Henning Kamp (phk@data.fls.dk).
 * 
 *   Ported to SunOS 5.x by John DiMarco (jdd@cdf.toronto.edu)
 */

#include <stdio.h>
#include <fcntl.h>

#ifdef FILENAME_MAX 	/* defined in stdio.h only in SYSV systems */
#define SYSV
#endif

#ifdef SYSV

#include <sys/scsi/generic/commands.h>
#include <sys/scsi/impl/types.h>
#include <sys/scsi/impl/uscsi.h>
/* Use sysinfo instead of gethostname to avoid having to link in libnsl */
#include <sys/systeminfo.h>
#define gethostname(b, l) sysinfo(SI_HOSTNAME, b, (long)l)

#else /* SYSV */

#include <scsi/impl/types.h>
#include <scsi/impl/uscsi.h>

#endif /* SYSV */

#define SUCCESS 0

unsigned char bdef[8000];
unsigned char bsav[8000];
unsigned char bcur[8000];
unsigned char bchg[8000];
int idef,isav,icur,ichg;
char hostname[100];

main(argc,argv)
    int argc;
    char **argv;
    {
    setbuf(stdout,0);
    gethostname(hostname,100);
    while(--argc)
        Do(*++argv);
    exit(0);
    }

HexDump(p,l,s)
    unsigned char *p;
    int l;
    char *s;
    {
    int i;

    while(l > 0)
	{
        printf("%s",s);
	for(i=0;i<16;i++)
	    {
	    if(i<l)
		printf("%02x ",p[i]);
	    else
		printf("   ");
	    }
	printf("   |");
	for(i=0;i<16;i++)
	    {
	    if(i<l && p[i] >= ' ' && p[i] <= '~')
		printf("%c",p[i]);
	    else
		printf(" ");
	    }
	printf("|\n");
	l-=16;
	p+=16;
	}

    }

Do(arg)
    char *arg;
    {
    int fd;
    int i,j;
    struct uscsi_cmd u,v,w;
    unsigned char buf[8000],*p,*q,*r;
    union scsi_cdb cdb;
    char *s;
    char f_ven[10];
    char f_typ[30];
    char f_rev[30];
    int f_trks_zone=-1;
    int f_asect=-1;
    int f_atrks=-1;
    int f_ncyl=-1;
    int f_pcyl=-1;
    int f_acyl=-1;
    int f_nhead=-1;
    int f_nsect=-1;
    int f_rpm=-1;
    int f_bpt=-1;
    int f_mgmt=-1;
    int f_capacity=-1;

    if((fd=open(arg,O_RDONLY))<SUCCESS)
	{
	perror(arg);
	return;
	}

    i=strlen(arg)+strlen(hostname)+1;
    while(i--)
	putchar('=');
    putchar('\n');
    printf("%s %s\n",hostname,arg);
    i=strlen(arg)+strlen(hostname)+1;
    while(i--)
	putchar('=');
    putchar('\n');
    
    u.uscsi_flags=USCSI_DIAGNOSE|USCSI_ISOLATE|USCSI_READ;
    u.uscsi_bufaddr=(caddr_t)buf;
    u.uscsi_buflen=8000;
    memset(buf,0,8000);
    cdb.scc_cmd = SCMD_READ_CAPACITY;
    cdb.scc_lun = 0;
    u.uscsi_cdb=(caddr_t)&cdb;
    u.uscsi_cdblen=10;
    i=ioctl(fd,USCSICMD,&u);
    if(i != SUCCESS)
	{
	fprintf(stderr,"i=%d ",i);
	perror("ioctl(READ_CAPACITY)");
	}
    printf("READ_CAPACITY:\n");
    printf("    Capacity:               %d\n",
        f_capacity=1+(buf[0]<<24)+(buf[1]<<16)+(buf[2]<<8)+buf[3]);

    u.uscsi_flags=USCSI_DIAGNOSE|USCSI_ISOLATE|USCSI_READ;
    u.uscsi_bufaddr=(caddr_t)buf;
    u.uscsi_buflen=8000;
    memset(buf,0,8000);
    cdb.scc_cmd = SCMD_INQUIRY;
    cdb.scc_lun = 0;
    cdb.g0_count0=200;
    u.uscsi_cdb=(caddr_t)&cdb;
    u.uscsi_cdblen=6;
    i=ioctl(fd,USCSICMD,&u);
    if(i != SUCCESS)
	{
	fprintf(stderr,"i=%d ",i);
	perror("ioctl(SCSI_INQUIRY)");
	}
    printf("INQUIRY:\n");
    printf("    Peripheral qualifier:   %d\n",(buf[0]>>5)&7);
    printf("    Peripheral device-type: %d\n",buf[0]&0x1f);
    printf("    Device-type modifier:   %d\n",buf[1]&0x7f);
    printf("    Removable medium:       %d\n",(buf[1]>>7)&0x1);
    printf("    ISO version:            %d\n",(buf[2]>>6)&0x3);
    printf("    ECMA version:           %d\n",(buf[2]>>3)&0x7);
    printf("    ANSI version:           %d\n",(buf[2]>>0)&0x7);
    printf("    AENC:                   %d\n",(buf[3]>>7)&0x1);
    printf("    TrmIOP:                 %d\n",(buf[3]>>6)&0x1);
    printf("    Response data format:   %d\n",(buf[3]>>0)&0xf);
    printf("    Relative addressing:    %d\n",(buf[7]>>7)&0x1);
    printf("    Wide bus 32:            %d\n",(buf[7]>>6)&0x1);
    printf("    Wide bus 16:            %d\n",(buf[7]>>5)&0x1);
    printf("    Sync transfer:          %d\n",(buf[7]>>4)&0x1);
    printf("    Linked commands:        %d\n",(buf[7]>>3)&0x1);
    printf("    Command queing:         %d\n",(buf[7]>>1)&0x1);
    printf("    Soft reset:             %d\n",(buf[7]>>0)&0x1);
    printf("    Vendor:                 '%-8.8s'\n",buf+8);
    printf("    Product:                '%-16.16s'\n",buf+16);
    printf("    Revision:               '%-4.4s'\n",buf+32);
    printf("    Comment1:               '%-20.20s'\n",buf+36);
    printf("    Comment2:               '%-40.40s'\n",buf+56);
    printf("    Comment3:               '%-40.40s'\n",buf+96);
    strncpy(f_ven,buf+8,8);
    for(s=f_ven+8-1;*s==' ';s--)
        *s=0;
    strncpy(f_typ,buf+16,16);
    for(s=f_typ+16-1;*s==' ';s--)
        *s=0;
    strncpy(f_rev,buf+32,4);
    for(s=f_rev+16-1;*s==' ';s--)
        *s=0;

    for(j=0;j<0x100;j+=0x40)
	{
	u.uscsi_flags=USCSI_DIAGNOSE|USCSI_ISOLATE|USCSI_READ;
	if(j==0x00) u.uscsi_bufaddr=(caddr_t)bcur;
	if(j==0x40) u.uscsi_bufaddr=(caddr_t)bchg;
	if(j==0x80) u.uscsi_bufaddr=(caddr_t)bdef;
	if(j==0xc0) u.uscsi_bufaddr=(caddr_t)bsav;
	u.uscsi_buflen=8000;
	memset(u.uscsi_bufaddr,0,8000);
	cdb.scc_cmd = SCMD_MODE_SENSE;
	cdb.scc_lun = 0;
	cdb.g0_addr1=0x3f | j;
	cdb.g0_count0=255;
	u.uscsi_cdb=(caddr_t)&cdb;
	u.uscsi_cdblen=6;
	i=ioctl(fd,USCSICMD,&u);
	if(i != SUCCESS)
	    {
	    fprintf(stderr,"i=%d ",i);
	    perror("ioctl(SCSI_MODE_SENSE)");
	    }
	}

#define IFDIFF(idx,val)	if(bsav[idx] != val || bcur[idx] != val || bchg[idx] != val || bdef[idx] != val)
#define IFUNEQ(idx)	if(bsav[idx] != bcur[idx] || bsav[idx] != bchg[idx] || bsav[idx] != bdef[idx])

	isav=icur=ichg=idef=0;
	IFUNEQ(0)
	    { printf("unequal page length in MODE SENSE PAGE\n"); return; }
        Pphead("*** Mode Sense:");
	Pp1("Medium Type","%d",1);
	Pp1("Device-specific","%d",2);
	Pp1("Length of pages(@8)","%d",3);
	isav=icur=ichg=idef=4;
	IFDIFF(3,8)
	    { printf("more than one page in MODE SENSE PAGE\n"); return; }
	Pp1("Density code","%d",0);
	Pp3("Number of blocks","%d",1);
	Pp3("Block length","%d",5);
	isav=icur=ichg=idef=12;
	p=bdef+idef;
	while(p < bdef+*bdef)
	    {
	    isav=icur=ichg=idef=p-bdef;
	    sprintf(buf,"*** Page %02x:",*p&0x3f);
	    Pphead(buf);
	    IFUNEQ(isav+1)
		{ printf("unequal page length in MODE SENSE PAGE\n"); return; }
	    switch(*p & 0x3f)
		{
		case 0x03:
		    f_trks_zone = Pp2("Tracks/zone","%d",2);
		    f_asect = Pp2("AltSect/zone","%d",4);
		    Pp2("AltTrack/zone","%d",6);
		    f_atrks = Pp2("AltTrack/LU","%d",8);
		    f_nsect = Pp2("Sect/track","%d",10);
		    Pp2("Byte/sect","%d",12);
		    Pp2("InterLeave","%d",14);
		    Pp2("Track skew","%d",16);
		    Pp2("Cyl. skew","%d",18);
		    break;
		case 0x04:
		    f_pcyl = Pp3("Cylinders","%d",2);
		    f_nhead = Pp1("Head","%d",5);
		    Pp3("Precomp cyl","%d",6);
		    Pp3("reduced cyl","%d",9);
		    Pp2("Step rate","%d",12);
		    Pp2("Landing zone","%d",14);
		    Pp1("Rotation offset","%d",18);
		    f_rpm = Pp2("RPM","%d",20);
		    break;
		}
	    HexDump(bcur+icur,bcur[1+icur]+2,"    curr |");
	    HexDump(bchg+ichg,bchg[1+ichg]+2,"    chg  |");
	    HexDump(bdef+idef,bdef[1+idef]+2,"    def  |");
	    HexDump(bsav+isav,bsav[1+isav]+2,"    save |");
	    p+=p[1]+2;
	    }


	if(f_rpm < 3000 || f_rpm > 10000)
	    f_rpm = 3600;

	if(f_asect*2 > f_nhead)
	    f_nsect--;
	for(i=0 ; i < 2 ; i++,f_nsect--)
	    {
	    f_acyl = f_atrks/f_nhead;
	    f_ncyl = f_capacity / f_nhead ;
	    f_ncyl /= f_nsect ;
	    f_ncyl -= f_acyl;

	    printf("\n\nSuggested format.dat based on default values:\n\n");
	    
	    printf("disk_type = \"something\" \\\n");
	    printf("\t: ctlr = SCSI : fmt_time = 4 \\\n");
	    printf("\t: rpm = %d : bpt = 40000 \\\n", f_rpm);
	    printf( "\t: nsect = %d : ncyl = %d : pcyl = %d : nhead = %d \\\n",
		f_nsect, f_ncyl, f_pcyl, f_nhead);
	    printf("\t: acyl = %d : atrks = %d : trks_zone = %d : asect = %d\n",
		f_acyl, f_atrks, f_trks_zone, f_asect);

	    printf("\npartition = \"something\" : disk = \"something\" : ctlr \"SCSI\" \\\n");
	    printf("\t: c = 0, %8d\n",f_ncyl*f_nsect*f_nhead);
	    }
	  
	return;
	}
#if 0
	{
	{
#define p2(x) ((p[x]<<8)+p[x+1])
#define p3(x) ((p[x]<<16)+(p[x+1]<<8)+p[x+2])

	while(p < buf+*buf)
	    {
	    printf("  Page %02x:\n",*p&0x3f);
	    switch(*p & 0x3f)
		{
		case 0x01:
		    printf("    AWRE:                   %d\n",p[2]&0x80?1:0);
		    printf("    ARRE:                   %d\n",p[2]&0x40?1:0);
		    printf("    TB:                     %d\n",p[2]&0x20?1:0);
		    printf("    RC:                     %d\n",p[2]&0x10?1:0);
		    printf("    ECC:                    %d\n",p[2]&0x08?1:0);
		    printf("    PER:                    %d\n",p[2]&0x04?1:0);
		    printf("    DTE:                    %d\n",p[2]&0x02?1:0);
		    printf("    DCR:                    %d\n",p[2]&0x01?1:0);
		    printf("    Read retry count:       %d\n",p[3]);
		    printf("    Correction span:        %d\n",p[4]);
		    printf("    Head offset count:      %d\n",p[5]);
		    printf("    Data strobe off. cnt:   %d\n",p[6]);
		    printf("    Write retry count:      %d\n",p[8]);
		    printf("    Recovery time limit:    %d\n",p2(10));
		    break;
		case 0x02:
		    printf("    Buffer full ratio:      %d\n",p[2]);
		    printf("    Buffer empty ratio:     %d\n",p[3]);
		    printf("    Bus inactivity limit:   %d\n",p2(4));
		    printf("    Disc. time limit:       %d\n",p2(6));
		    printf("    Conn. time limit:       %d\n",p2(8));
		    printf("    Max burst size:         %d\n",p2(10));
		    printf("    DTDC:                   %d\n",p[12]&3);
		    break;
		case 0x03:
		    printf("    Tracks/zone:            %d\n",
			f_trks_zone=(p[2]<<8)+p[3]);
		    printf("    AltSect/zone:           %d\n",
			f_asect=(p[4]<<8)+p[5]);
		    printf("    AltTrack/zone:          %d\n",(p[6]<<8)+p[7]);
		    printf("    AltTrack/LU:            %d\n",
			f_atrks=(p[8]<<8)+p[9]);
		    printf("    Sect/track:             %d\n",
			f_nsect=(p[10]<<8)+p[11]);
		    printf("    Byte/sect:              %d\n",(p[12]<<8)+p[13]);
		    printf("    Interleave:             %d\n",(p[14]<<8)+p[15]);
		    printf("    Track skew:             %d\n",(p[16]<<8)+p[17]);
		    printf("    Cylinder skew:          %d\n",(p[18]<<8)+p[19]);
		    printf("    SSEC:                   %d\n",p[20]&0x80?1:0);
		    printf("    HSEC:                   %d\n",p[20]&0x40?1:0);
		    printf("    RMB:                    %d\n",p[20]&0x20?1:0);
		    printf("    SURF:                   %d\n",p[20]&0x10?1:0);
		    break;
		case 0x04:
		    printf("    Cylinders:              %d\n",
			f_pcyl=(p[2]<<16)+(p[3]<<8)+p[4]);
		    printf("    Heads:                  %d\n",
			f_nhead=p[5]);
		    printf("    Write precomp start:    %d\n",p2(6));
		    printf("    Write reduced start:    %d\n",p2(8));
		    printf("    Drive step rate:        %d\n",p2(10));
		    printf("    Landing zone:           %d\n",p2(12));
		    printf("    Rotation pos. lock:     %d\n",(p[17]>>0) & 0x3);
		    printf("    Rotation offset:        %d\n",p[18]);
		    printf("    RPM:                    %d\n",f_rpm=p2(20));
		    break;
		case 0x07:
		    printf("    ERR:                    %d\n",(p[2]>>3) & 0x1);
		    printf("    PER:                    %d\n",(p[2]>>2) & 0x1);
		    printf("    DTE:                    %d\n",(p[2]>>1) & 0x1);
		    printf("    DCR:                    %d\n",(p[2]>>0) & 0x1);
		    printf("    Verify retry count:     %d\n",p[3]);
		    printf("    Verify retry span:      %d\n",p[4]);
		    printf("    Recovery time limit:    %d\n",p2(10));
		    break;
		case 0x08:
		    printf("    Write cache enable:     %d\n",(p[2]>>2) & 0x1);
		    printf("    Multiplication factor:  %d\n",(p[2]>>1) & 0x1);
		    printf("    Read cache disable:     %d\n",(p[2]>>0) & 0x1);
		    printf("    Demand rd reten. pri:   %d\n",(p[3]>>4)&0x0f);
		    printf("    Write reten. pri:       %d\n",(p[3]>>0)&0x0f);
		    printf("    Dis. pre-fetch xfer-len:%d\n",p2(4));
		    printf("    Minimum pre-fetch:      %d\n",p2(6));
		    printf("    Maximum pre-fetch:      %d\n",p2(8));
		    printf("    Max pre-fetch-ceil:     %d\n",p2(10));
		    break;
		case 0x09:
		    printf("    Interface Identifier:   %d\n",p2(2));
		    break;
		case 0x0a:
		    printf("    RLEC:                   %d\n",p[2] & 0x1);
		    printf("    Queue Alg. modifier:    %d\n",(p[3]>>4) & 0xf);
		    printf("    Queue Err. mgmt:        %d\n",(p[3]>>1) & 0x1);
		    printf("    Disable queing:         %d\n",(p[3]>>0) & 0x1);
		    printf("    EECA:                   %d\n",(p[4]>>7) & 0x1);
		    printf("    RAENP:                  %d\n",(p[4]>>2) & 0x1);
		    printf("    UAAENP:                 %d\n",(p[4]>>1) & 0x1);
		    printf("    EAENP:                  %d\n",(p[4]>>0) & 0x1);
		    printf("    EAN Holdoff:            %d\n",(p[6]<<8)+p[7]);
		    break;
		case 0x0b:
		    printf("    Medium type 1:          %d\n",p[4]);
		    printf("    Medium type 2:          %d\n",p[5]);
		    printf("    Medium type 3:          %d\n",p[6]);
		    printf("    Medium type 4:          %d\n",p[7]);
		    break;
		case 0x0c:
		    printf("    Notched drive:          %d\n",(p[2]>>7) & 0x1);
		    printf("    Log/Phys notch:         %d\n",(p[2]>>6) & 0x1);
		    printf("    Max notches:            %d\n",p2(4));
		    printf("    Active notch:           %d\n",p2(6));
		    printf("    Start boundary:         %d\n",p2(8));
		    printf("    End boundary:           %d\n",p2(10));
		    printf("    Pages notched:          0x%04x\n",p2(12));
		    break;
		default:
		    HexDump(p,p[1]+2);
		    break;
		}
	    p+=p[1]+2;
	    }
	}
    if(f_capacity != -1)
	{
	f_ncyl=f_capacity/(f_trks_zone*f_nsect-f_asect);
	}
    else
	{
	f_ncyl=f_pcyl-f_mgmt-f_atrks/f_nhead;
	}
    printf("\014");
    printf("# Vendor   '%s'\n",f_ven);
    printf("# Type     '%s'\n",f_typ);
    printf("# Revision '%s'\n",f_rev);
    printf("disk_type = \"%s %s (%s)\" \\\n",f_ven,f_typ,f_rev);
    printf("\t: ctlr = MD21 \\\n");
    printf("\t: pcyl = %d \\\n",f_pcyl);
    printf("\t: ncyl = %d \\\n",f_ncyl);
    printf("\t: nhead = %d \\\n",f_nhead);
    printf("\t: nsect = %d \\\n",f_nsect);
    printf("\t: atrks = %d \\\n",f_atrks);
    printf("\t: asect = %d \\\n",f_asect);
    printf("\t: trks_zone = %d \\\n",f_trks_zone);
    printf("\t: rpm = %d \\\n",3600);
    printf("\t: bpt = %d \n",36000);
    printf("\n");
    printf("# One zone = %d cyl\n",f_trks_zone/f_nhead);
    printf("# One zone = %d blocks\n",(f_trks_zone*f_nsect-f_asect));
    printf("# One zone = %d kB\n",(f_trks_zone*f_nsect-f_asect)/2);
    printf("partition = \"%s %s (%s)\" \\\n",f_ven,f_typ,f_rev);
    printf("\t: disk = \"%s %s (%s)\" \\\n",f_ven,f_typ,f_rev);
    printf("\t: ctlr = MD21 \\\n");
    printf("\t: c = 0, %d\n",
        (f_ncyl*f_nhead)*(f_trks_zone*f_nsect-f_asect)/f_trks_zone);
    }
#endif

#define q2(p,x) ((p[x]<<8)+p[x+1])
#define q3(p,x) ((p[x]<<16)+(p[x+1]<<8)+p[x+2])
Pphead(s)
    char *s;
    {
    printf("\n%-20s",s);
    printf("%15s","curr");
    printf("%15s","chg");
    printf("%15s","def");
    printf("%15s\n","save");
    }

Pp1(ldr,fmt,idx)
    char *ldr,*fmt;
    int idx;
    {
    char s[20];
    printf("%-20s",ldr);
    sprintf(s,fmt,bcur[idx+icur]); printf("%15s",s);
    sprintf(s,fmt,bchg[idx+ichg]); printf("%15s",s);
    sprintf(s,fmt,bdef[idx+idef]); printf("%15s",s);
    sprintf(s,fmt,bsav[idx+isav]); printf("%15s\n",s);
    return bdef[idx+idef];
    }

Pp2(ldr,fmt,idx)
    char *ldr,*fmt;
    int idx;
    {
    char s[20];
    printf("%-20s",ldr);
    sprintf(s,fmt,q2(bcur,idx+icur)); printf("%15s",s);
    sprintf(s,fmt,q2(bchg,idx+ichg)); printf("%15s",s);
    sprintf(s,fmt,q2(bdef,idx+idef)); printf("%15s",s);
    sprintf(s,fmt,q2(bsav,idx+isav)); printf("%15s\n",s);
    return q2(bdef,idx+idef);
    }

Pp3(ldr,fmt,idx)
    char *ldr,*fmt;
    int idx;
    {
    char s[20];
    printf("%-20s",ldr);
    sprintf(s,fmt,q3(bcur,idx+icur)); printf("%15s",s);
    sprintf(s,fmt,q3(bchg,idx+ichg)); printf("%15s",s);
    sprintf(s,fmt,q3(bdef,idx+idef)); printf("%15s",s);
    sprintf(s,fmt,q3(bsav,idx+isav)); printf("%15s\n",s);
    return q3(bdef,idx+idef);
    }

